----------------------------------------------------------------------------------
-- Create Date:    14:31:43 05/15/2012 
----------------------------------------------------------------------------------
library ieee;
library work;

use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.config.all;

entity centroid is
    Port (
        clk       : in  std_logic;
        reset     : in  std_logic;
        dowork    : in  std_logic; 
        num_ref   : in  integer range 0 to MAX_FBGS; 
        num_sen   : in  integer range 0 to MAX_FBGS; 
        init      : in  std_logic;
        mid       : in  std_logic;
        val       : in  std_logic_vector(RAMP_BITS-1 downto 0); 
        amplitude : in  unsigned (AMPL_BITS-1 downto 0);
        adc_done  : in  std_logic;
        peak      : out std_logic_vector(RAMP_BITS-1 downto 0));
end centroid;

architecture Behavioral of centroid is
    constant NUM_BITS : natural := RAMP_BITS+RAMP_BITS+AMPL_BITS;
    constant DEN_BITS : natural := RAMP_BITS+AMPL_BITS;
    
    -- signals for restoring divisor
    -- atention for the bits of numerator, denumerator and quotient
    signal numerator_reg, numerator_next       : unsigned(NUM_BITS-1 downto 0); 
    signal denumerator_reg, denumerator_next   : unsigned(DEN_BITS-1 downto 0); 
    signal restart_reg, restart_next, div_done : std_logic; 
    signal quotient                            : unsigned(RAMP_BITS-1 downto 0);

    type state_type is (idle, outside, inside, cal_valve, div_pre, dividing);  
    signal state_reg , state_next  : state_type; 
    signal sum0_reg, sum0_next: unsigned(NUM_BITS-1 downto 0); 
    signal sum1_reg, sum1_next: unsigned(DEN_BITS-1 downto 0); 

    signal peak_reg, peak_next: unsigned(RAMP_BITS-1 downto 0); -- difference between sensor peak and reference peak
    signal out_reg , out_next : std_logic_vector(RAMP_BITS-1 downto 0); -- data type conversion 

    signal position : unsigned(RAMP_BITS-1 downto 0); 
    signal sum_amp_reg, sum_amp_next: unsigned(DEN_BITS-1 downto 0); 
    signal valve_reg, valve_next: unsigned(AMPL_BITS-1 downto 0);
    signal valve_up_reg, valve_up_next: unsigned(AMPL_BITS-1 downto 0); 
    
    signal num_peak_reg, num_peak_next: integer range 0 to MAX_FBGS; 
    signal peak_ref_reg, peak_ref_next: unsigned(RAMP_BITS-1 downto 0); 
    
begin

position <= unsigned(val); 


unit_dividing: entity work.restoring(behavioral)
generic map(num_bits => NUM_BITS, den_bits => DEN_BITS, quot_bits => RAMP_BITS)
port map (clk => clk, reset => reset, numerator => numerator_reg, 
          denumerator => denumerator_reg, restart => restart_reg, 
          quotient => quotient, div_done => div_done );
          
--FSMD register
process(clk, reset)
begin
    if rising_edge(clk) then
        if reset = '1' or dowork = '0' then
            state_reg       <= idle; 
            sum0_reg        <= (others => '0'); 
            sum1_reg        <= (others => '0'); 
            numerator_reg   <= (others => '0'); 
            denumerator_reg <= (others => '0'); 
            restart_reg     <= '1'; 
            sum_amp_reg     <= (others => '0'); 
            valve_reg       <= (others => '0'); 
            valve_up_reg    <= (others => '0'); 
            peak_reg        <= (others => '0'); 
            out_reg         <= (others => '0'); 
            num_peak_reg    <= 0; 
            peak_ref_reg    <= (others => '0'); 
        else
            state_reg       <= state_next; 
            sum0_reg        <= sum0_next; 
            sum1_reg        <= sum1_next; 
            numerator_reg   <= numerator_next; 
            denumerator_reg <= denumerator_next; 
            restart_reg     <= restart_next; 
            sum_amp_reg     <= sum_amp_next; 
            valve_reg       <= valve_next;
            valve_up_reg    <= valve_up_next; 
            peak_reg        <= peak_next; 
            out_reg         <= out_next;
            num_peak_reg    <= num_peak_next; 
            peak_ref_reg    <= peak_ref_next; 
        end if; 
     end if; 
end process; 

-- FSMD next logic
process(init, mid, amplitude, adc_done, state_reg, sum0_reg, sum1_reg, 
        out_reg, restart_reg, numerator_reg, denumerator_reg,
        position, div_done, quotient, sum_amp_reg, valve_reg, valve_up_reg, peak_reg, num_peak_reg, 
        peak_ref_reg, num_sen, num_ref)
variable level : unsigned(AMPL_BITS-1 downto 0);        
begin 
    state_next       <= state_reg; 
    sum0_next        <= sum0_reg; 
    sum1_next        <= sum1_reg; 
    restart_next     <= '1';
    numerator_next   <= numerator_reg; 
    denumerator_next <= denumerator_reg; 
    sum_amp_next     <= sum_amp_reg; 
    valve_next       <= valve_reg; 
    valve_up_next    <= valve_up_reg; 
    peak_next        <= peak_reg; 
    out_next         <= out_reg;
    num_peak_next    <= num_peak_reg; 
    peak_ref_next    <= peak_ref_reg; 
    
    case state_reg is
        when idle =>
            if init = '1' then
                sum0_next     <= (others => '0'); 
                sum1_next     <= (others => '0'); 
                sum_amp_next  <= (others => '0');
                peak_next     <= (others => '0');
                state_next    <= outside; 
                num_peak_next <= 0; 
                peak_ref_next <= (others => '0'); 
            end if; 
            
        when outside =>
            if mid = '0' then
                if adc_done = '1' then
                    -- automatic mean level control
                    if amplitude >= valve_reg then
                        level := amplitude - valve_reg;
                    else
                        level := to_unsigned(0, AMPL_BITS);
                    end if;
                    
                    sum0_next    <= sum0_reg + level*position; 
                    sum1_next    <= sum1_reg + level; 
                    sum_amp_next <= sum_amp_reg + amplitude; 
                    
                    if amplitude >= valve_up_reg then
                        state_next <= inside; 
                    end if;   
                 end if; 
            else
                state_next <= cal_valve;     
            end if;
            
         when inside =>
            if mid = '0' then
                  if adc_done = '1' then
                    -- automatic mean level control
                    if amplitude >= valve_reg then
                        level := amplitude - valve_reg;
                    else
                        level := to_unsigned(0, AMPL_BITS);
                    end if;
                    
                    sum0_next    <= sum0_reg + level*position; 
                    sum1_next    <= sum1_reg + level; 
                    sum_amp_next <= sum_amp_reg + amplitude; 
                    
                    if amplitude <= valve_reg then
                        num_peak_next <= num_peak_reg + 1; 
                        state_next <= div_pre; 
                    end if; 
                  
                  end if;       
            else
                 state_next <= cal_valve;
            end if;
            
         when cal_valve =>
            if num_peak_reg < num_sen then   -- wrong sensor number input detection
                out_next  <= (others => '1'); -- the difference between sensor and reference FBG should not be 4095
            else 
                out_next      <= std_logic_vector(peak_reg);
            end if; 
            valve_next    <= sum_amp_reg(DEN_BITS-1 downto RAMP_BITS) + to_unsigned(VALVE_RISE, AMPL_BITS);
            valve_up_next <= sum_amp_reg(DEN_BITS-1 downto RAMP_BITS) + to_unsigned(PEAK_UP_THRES, AMPL_BITS);
            state_next    <= idle; 
            
         when div_pre =>
            numerator_next   <= sum0_reg; 
            denumerator_next <= sum1_reg;
            sum0_next        <= (others => '0'); 
            sum1_next        <= (others => '0'); 
            state_next       <= dividing; 
            
         when dividing =>
            restart_next <= '0'; 
            if div_done = '1' then
--                peak_next  <= quotient - peak_reg; 
                if num_ref = 0 then
                    peak_ref_next <= (others => '0'); 
                else
                    if num_peak_reg = num_ref then
                    peak_ref_next <= quotient; 
                    end if; 
                end if; 
                
                if num_sen = 0 then
                    peak_next <= (others => '0'); 
                else
                    if num_peak_reg = num_sen then
                        peak_next <= quotient - peak_ref_reg; 
                    end if; 
                end if; 
                
                state_next <= outside; 
            end if; 
    end case; 
end process; 
            
-- FSMD output
peak <= out_reg; 

end Behavioral;