----------------------------------------------
-- Create Date:    15:07:11 06/27/2012 
----------------------------------------------
library IEEE;
library work; 

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.config.all; 

entity moving_average is
    Port ( 
        clk       : in  std_logic;
        reset     : in  std_logic;
        dowork    : in  std_logic; 
        num_ref   : in  integer range 0 to MAX_FBGS; 
        num_sen   : in  integer range 0 to MAX_FBGS; 
        init      : in  std_logic;
        mid       : in  std_logic;
        amplitude : in  unsigned(AMPL_BITS-1 downto 0);
        adc_done  : in  std_logic;
        peak      : out std_logic_vector(RAMP_BITS-1 downto 0)
    );
end moving_average;

architecture Behavioral of moving_average is
    constant SUM_WIND_BITS  : natural := AMPL_BITS + WIND_LEN_BITS; 
    constant SUM_AMPL_BITS  : natural := AMPL_BITS + RAMP_BITS; 
    constant WIND_LEN       : natural := 2**WIND_LEN_BITS; 
    constant RAMP_LEN       : natural := 2**RAMP_BITS; 
    constant LEFT           : natural := 2**(WIND_LEN_BITS-1); 
    
    --Signals for FSMD
    type state_type is (idle, store, initialization, filtering, mean, inside, outside,
        initialization2, filtering2, mean2, inside2, outside2,
        initialization3, filtering3, mean3, inside3, outside3, peak_local); 
    signal state           : state_type; 
    signal sum_win         : unsigned(SUM_WIND_BITS-1 downto 0); 
    signal sum_amp         : unsigned(SUM_AMPL_BITS-1 downto 0); 
    signal valve, valve_up : unsigned(AMPL_BITS-1 downto 0);
    signal pamp_local      : unsigned(AMPL_BITS-1 downto 0); 
    signal pt_local        : unsigned(RAMP_BITS-1 downto 0); 
    signal dif             : std_logic_vector(RAMP_BITS-1 downto 0);
    signal num_peak: integer range 0 to MAX_FBGS; -- detect various FBGs
    signal peak_ref: unsigned(RAMP_BITS-1 downto 0); -- registrate the reference FBG peak
    
    --Memory for storing the specter data
    signal wr_spectra, rd_spectra          : std_logic;
    signal wr_spectra_data, rd_spectra_data: unsigned(AMPL_BITS-1 downto 0);
    signal spectra_addr                    : integer range 0 to RAMP_LEN-1;
    
    --Memory for storing the filter window data
    signal wr_window, rd_window            : std_logic;
    signal wr_window_data, rd_window_data  : unsigned(AMPL_BITS-1 downto 0);
    signal window_addr                     : integer range 0 to WIND_LEN-1;
    
    -- counters
    signal t               : natural range 0 to RAMP_LEN;
    signal win_t           : natural range 0 to WIND_LEN-1;
    signal fil_t           : natural range 0 to FIL_NUM-1;
begin

unit_mem_spectra: entity work.mem_block(behavioral)
generic map(len => RAMP_LEN, bits => AMPL_BITS)
port map (clk => clk, wr => wr_spectra, rd => rd_spectra,
    addr => spectra_addr, rd_data => rd_spectra_data, wr_data => wr_spectra_data
);

unit_mem_window: entity work.mem_block(behavioral)
generic map(len => WIND_LEN, bits => AMPL_BITS)
port map (clk => clk, wr => wr_window, rd => rd_window,
    addr => window_addr,rd_data => rd_window_data, wr_data => wr_window_data
);

process(clk, reset)
begin
    if rising_edge(clk) then
        if reset = '1' or dowork = '0' then
            state      <= idle; 
            t          <= 0; 
            win_t      <= 0;
            fil_t      <= 0;
            sum_amp    <= (others => '0');
            sum_win    <= (others => '0');
            valve      <= (others => '0'); 
            valve_up   <= (others => '0'); 
            pamp_local <= (others => '0'); 
            pt_local   <= (others => '0');
            dif        <= (others => '0'); 
            num_peak   <= 0; 
            peak_ref   <= (others => '0'); 
        else
            wr_spectra <= '0';
            rd_spectra <= '0'; 
            
            wr_window  <= '0';
            rd_window  <= '0';
            
            case state is
                when idle =>
                    if init = '1' then
                        t     <= 0;
                        state <= store;
                        num_peak <= 0; 
                        peak_ref <= (others => '0'); 
                    end if;
            
                when store =>
                    if mid = '1' then
                        t       <= 0;
                        fil_t   <= 0;
                        sum_win <= (others => '0');
                        state   <= initialization;
                    else
                        if adc_done = '1' then
                            spectra_addr    <= t;
                            wr_spectra_data <= amplitude;
                            wr_spectra      <= '1';
                            t               <= t + 1;
                        end if;
                    end if;
                
                when initialization =>
                    if t = WIND_LEN then
                        win_t   <= 0;
                        state   <= filtering;
                    else
                        spectra_addr <= t;
                        rd_spectra   <= '1';
                        state        <= initialization2;
                    end if;
                    
                when initialization2 =>
                    state <= initialization3;
                    
                when initialization3 =>
                    sum_win         <= sum_win + rd_spectra_data;
                    window_addr     <= t;
                    wr_window_data  <= rd_spectra_data;
                    wr_window       <= '1';
                    t               <= t + 1;
                    state           <= initialization;
                
                when filtering =>
                    if t = RAMP_LEN then
                        if fil_t = FIL_NUM - 1 then
                            t       <= 0;
                            sum_amp <= (others => '0');
                            state   <= mean;
                        else
                            t       <= 0;
                            sum_win <= (others => '0');
                            fil_t   <= fil_t + 1;
                            state   <= initialization;
                        end if;
                    else
                        spectra_addr  <= t;
                        rd_spectra    <= '1';
                        
                        window_addr   <= win_t;
                        rd_window     <= '1';
                        
                        state         <= filtering2;
                    end if;
                    
                when filtering2 =>
                    state <= filtering3;
                
                when filtering3 =>
                    sum_win <= sum_win - rd_window_data + rd_spectra_data;
                    
                    spectra_addr    <= t - LEFT;
                    wr_spectra_data <= sum_win(SUM_WIND_BITS-1 downto WIND_LEN_BITS);
                    wr_spectra      <= '1';
                    
                    window_addr     <= win_t;
                    wr_window_data  <= rd_spectra_data;
                    wr_window       <= '1';
                    
                    if win_t = WIND_LEN-1 then win_t <= 0; else win_t <= win_t + 1; end if;
                    t <= t + 1;
                    state <= filtering;
                
                when mean =>
                    if t = RAMP_LEN then
                        valve    <= sum_amp(SUM_AMPL_BITS-1 downto RAMP_BITS) + to_unsigned(2, AMPL_BITS);
                        valve_up <= sum_amp(SUM_AMPL_BITS-1 downto RAMP_BITS) + to_unsigned(PEAK_UP_THRES, AMPL_BITS);
                        t        <= 0;
                        pt_local <= (others => '0');
                        state    <= outside;
                    else
                        spectra_addr <= t;
                        rd_spectra   <= '1';
                        state        <= mean2;
                    end if;
                
                when mean2 =>
                    state    <= mean3;
                
                when mean3 =>
                    sum_amp  <= sum_amp + rd_spectra_data;
                    t        <= t + 1;
                    state    <= mean;
                    
                when outside =>
                    if t = RAMP_LEN then
                        if num_peak < num_sen then   -- wrong sensor number input detection
                            dif  <= (others => '1'); -- the difference between sensor and reference FBG should not be 4095
                        end if; 
                        state <= idle;
                    else
                        spectra_addr <= t;
                        rd_spectra   <= '1';
                        state        <= outside2;
                    end if;
                
                when outside2 =>
                    state <= outside3;
                
                when outside3 =>
                    if rd_spectra_data >= valve_up then
                        pamp_local <= valve;
                        state      <= inside;
                    else
                        t <= t + 1;
                        state <= outside;
                    end if;
                  
                when inside =>
                    if t = RAMP_LEN then
                        if num_peak < num_sen then   -- wrong sensor number input detection
                            dif  <= (others => '1'); -- the difference between sensor and reference FBG should not be 4095
                        end if; 
                        state <= idle;
                    else
                        spectra_addr <= t;
                        rd_spectra   <= '1';
                        state        <= inside2;
                    end if;
                
                when inside2 =>
                    state <= inside3;
                    
                when inside3 =>
                    if rd_spectra_data <= valve then
                        state <= peak_local;
                        num_peak <= num_peak + 1; 
                    else
                        if rd_spectra_data > pamp_local then
                            pamp_local  <= rd_spectra_data;
                            pt_local    <= to_unsigned(t, RAMP_BITS);
                        end if;
                            
                        t <= t + 1;
                        state <= inside;
                    end if;
                    
                when peak_local =>
                    if num_ref = 0 then
                    peak_ref <= (others => '0'); 
                    else
                        if num_peak= num_ref then
                            peak_ref<= pt_local; 
                        end if; 
                    end if; 
                
                    if num_sen = 0 then
                        dif <= (others => '0'); 
                    else
                        if num_peak = num_sen then
                            dif <= std_logic_vector(pt_local - peak_ref); 
                        end if; 
                    end if; 
                    state <= outside; 
                    
            end case;            
        end if;
    end if;
end process;

-- output logic
peak <= dif;

end Behavioral;